/*
*  C Implementation: AVR ISP Programmer 
*
* Description:
*
*  A serial programmer  somewhat compatible with "butterfly" "avr109" and
*     "avr910" modes of avrdude
*
*
* Author: Koji Suzuki (C) 2007
* ATTiny2313_bootloader Author:
*              Ernst Bachmann <e.bachmann@netandweb.de>, (C) 2006
*
* Copyright: See License.txt file that comes with this distribution
*
*/

#include "usb910conf.h"

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>
#include <util/delay.h>

#include "usbcdc.h"
#include "usb910.h"

#if !defined(NO_AVR910) && ( TX_SIZE < (32) )
#error "TX_SIZE is too small"
#endif

#ifdef JTAG_EXTENSION
#define JTAG_BUFSIZE	64
#endif

#define getc	usbcdc_getc
#define putc	usbcdc_putc
#define poll	usbcdc_poll

#define bit_set(port,bit)	{ port |= (1<<(bit));}
#define bit_clear(port,bit)	{ port &= ~(1<<(bit));}
#define min(a,b)		((a)<(b) ? (a):(b))

#define swap(v) __asm__ volatile ("swap %0" : "=r" (v) : "0" (v) )
#define nop()   __asm__ volatile ("nop" )

union {
	uint16_t word;
	struct {
		uint8_t lo;
		uint8_t hi;
	};
} rate,sample_num;

static inline void set_rate() {
#ifdef __AVR_ATtiny861__
	TCCR1B = 0;
	TCCR0B = 0;
	TCNT1 = 0;
	TCNT0L = 0;
	TCNT0H = 0;

	switch (rate.hi & 0x7) { // 1: 1/1 , 2:1/8, 3:1/64, 4:1/256, 5:1/1024
	case 1: TCCR1B =  1;  break;
	case 2: TCCR1B =  3;  break;
	case 3: TCCR1B =  7;  break;
	case 4: TCCR1B =  9;  break;
	default:
	case 5: TCCR1B = 11;  break;
	}
	OCR1C = rate.lo;
	TCCR0B = 1; // free running @ 1/1
#else
	TCCR0B = 0;
	TCCR1B = 0;
	TCNT0 = 0;
	TCNT1 = 0L;

	TCCR0A = (2<<WGM00);
	TCCR0B = (rate.hi & 0x7); // 1: 1/1 , 2:1/8, 3:1/64, 4:1/256, 5:1/1024
	OCR0A = rate.lo;
	TCCR1B = 1; // free running @ 1/1
#endif
}


#define nop()	__asm__ volatile ("nop" )

#define getc	usbcdc_getc
#define putc	usbcdc_putc
#define poll	usbcdc_poll

int main(void)
{
uint8_t ch;

    /* set system_clock pre scaler 1/8 => 1/1 */
    CLKPR = (1<<CLKPCE);
    CLKPR = 0;

    wdt_enable(WDTO_1S);
    usbcdc_init();

    sei();

    for(;;) {
	    wdt_reset();
	    poll();
	    ch = getc();
	    if (ch == 'S') {
		    putc('U');
		    putc('S');
		    putc('B');
		    putc('1');
		    putc('6');
		    putc('2');
		    putc('B');
	    } else if (ch == 'p') { // get Programmer Type
		   putc('S'); // Serial
	    } else if ((ch == 'V') || (ch == 'v')) {
		    putc('1');
		    putc('0');
	    } else if ((ch == 0x1b) || (ch == '\n')) { // ESC,\n: nop
	    } else if (ch == 'r') { // set rate
		    rate.lo = getc();
		    rate.hi = getc();
		    putc('\r'); // ACK
	    } else if (ch == 's') { // set samples
		    sample_num.lo = getc();
		    sample_num.hi = getc();
		    putc('\r'); // ACK
	    } else if (ch == 'g') { // go 
		    set_rate();
		    while (sample_num.word) {
	   		    wdt_reset();
#ifdef __AVR_ATtiny861__
			    if (bit_is_set(TIFR, TOV1)) {
			    	bit_set(TIFR, TOV1);
				putc(TCNT0L);
				putc(TCNT0H);
				putc(TCNT0L);
				putc(TCNT0H);
		    		sample_num.word--;
			    }
#else
			    if (bit_is_set(TIFR0, OCF0A)) {
			    	bit_set(TIFR0, OCF0A);
				putc(TCNT1L);
				putc(TCNT1H);
				putc(TCNT1L);
				putc(TCNT1H);
		    		sample_num.word--;
			    }
#endif
		    }
	    } else if (ch == 'h') { // go + 1
		    set_rate();
		    while (sample_num.word) {
	   		    wdt_reset();
#ifdef __AVR_ATtiny861__
			    if (bit_is_set(TIFR, TOV1)) {
			    	bit_set(TIFR, TOV1);
				getc();
				getc();
				getc();
				getc();
		    		sample_num.word--;
			    }
#else
			    if (bit_is_set(TIFR0, OCF0A)) {
			    	bit_set(TIFR0, OCF0A);
				getc();
				getc();
				getc();
				getc();
		    		sample_num.word--;
			    }
#endif
		    }
	    } else if (ch == 'i') { // go + 2
		    set_rate();
		    while (sample_num.word) {
	   		    wdt_reset();
#ifdef __AVR_ATtiny861__
			    if (bit_is_set(TIFR, TOV1)) {
			    	bit_set(TIFR, TOV1);
				getc();
				getc();
				getc();
				getc();
				putc(TCNT0L);
				putc(TCNT0H);
				putc(TCNT0L);
				putc(TCNT0H);
		    		sample_num.word--;
			    }
#else
			    if (bit_is_set(TIFR0, OCF0A)) {
			    	bit_set(TIFR0, OCF0A);
				getc();
				getc();
				getc();
				getc();
				putc(TCNT1L);
				putc(TCNT1H);
				putc(TCNT1L);
				putc(TCNT1H);
		    		sample_num.word--;
			    }
#endif
		    }
	    } else {
		    putc('?');
	    }
    }
    return 0;
}
